@import url('https://fonts.googleapis.com/css?family=Fredoka+One');

div.sphinxsidebarwrapper p.logo {
    margin-bottom: 1.5em;
}


h1,
h2,
h3,
.admonition-title {
    font-family: 'Fredoka One', serif !important;
}

h1 {
    color: #b83f3f !important;
}

h2 {
    color: #3a76a8 !important;
}

h3 {
    color: #266f2b !important;
}

div.sphinxsidebar h3 a {
    color: #b79414 !important;
}


.admonition.warning,
.admonition.caution,
.admonition.tip {
    color: white;
    background-color: silver;
    border: none;
    border-radius: 6px;
    padding: 20px 30px;
}

.admonition.tip {
    background-color: #3a76a8;
}

.admonition.warning {
    background-color: #b83f3f;
}

.admonition.caution {
    background-color: #ffc238;
    color: black;
}

.admonition.warning a {
    color: red;
    border-bottom: none;
    text-decoration: underline;
}

.admonition.tip a {
    color: #00b8e6;
    border-bottom: none;
    text-decoration: underline;
}

.admonition code {
    background-color: rgba(255, 255, 255, 0.2) !important;
    border-bottom: none !important;
}

ol li {
    margin: 1em 0;
}

.pyweek .admonition-title {
    display: none;
}
